import {
  Component,
  ElementRef,
  ViewChild,
  OnInit,
  Input,
  OnDestroy,
  HostListener
  
} from "@angular/core";
import { UUID } from 'angular2-uuid';

import {
  MySprite,
  RichText,
  getMinScale,
  ShapeRect,
  ShapeCircle,
  tweenChange,
  randomSortByArr,
  Label,
  showPopParticle,
  moveItem,
  removeItemFromArr,
  rotateItem,
  hideItem,
  showItem,
  ShapeRectNew,
  scaleItem,
  showBlingBling,
  waterWave,
  jelly,
  getAngleByPos,
  getPosDistance
} from "./Unit";

import { localImages, localAudios } from "./resources";

import { Cartoon } from './Cartoon'
import { Subject } from "rxjs";
import { debounceTime, map, takeWhile, retry } from "rxjs/operators";
import * as _ from "lodash";
import TWEEN from "@tweenjs/tween.js";
import defauleFormData from '../../assets/play/default/formData/defaultData.js'
import { element } from 'protractor';


const zIndexMap = {
  mainBackground: 0,
  answerCard: 100,
  questionCard: 100,
  questionKuang: 50,
  imageContainer: 20,
  shortAudio: 60,
  progressBar: 55
}

@Component({
  selector: "app-play",
  templateUrl: "./play.component.html",
  styleUrls: ["./play.component.scss"]
})

export class PlayComponent implements OnInit, OnDestroy {
  
  g_cartoon = new Cartoon()

  // ------------  全局数据  ------------
  g_stage; //中心舞台
  // g_background_color = "#rgb(0,0,0,0)"
  g_background_color = "#af7cb7"
  g_enableMapDown = true; // 触摸使能
  g_enableMapUp = true; // 抬起使能
  g_enableMapMove = true; // 移动ss使能
  g_canvasLeft;
  g_canvasTop;
  g_animationId: any;
  g_mapScale = 1; // 缩放比例
  g_KEY = "DataKey_PU01";
  g_canvasWidth = 1280;
  g_canvasHeight = 720;
  g_canvasBaseW = 1280;
  g_canvasBaseH = 720;
  g_winResizeEventStream = new Subject();
  g_clickX; // 点击坐标 X
  g_clickY; // 点击坐标 Y
  g_ctx; // canvas 实例
  g_data; // 数据
  g_formData; // 核心表单数据
  g_teacherFlag = false; // 默认角色
  g_currentUser;
  g_firstTouch = true;
  g_partTitle_x = null;
  g_mainTitle_x = null;
  // ------------------------------------

  // ------------  私有数据  ------------
  m_mapDownQueue = {} //按下事件处理队列
  m_mapDownArray = [] //按下事件处理队列
  m_mapDownObject = []

  m_mapUpQueue = {} //抬起事件处理队列
  m_mapUpArray = [] //抬起事件处理队列
  m_mapUpObject = []

  m_mapMoveArray = [] //移动事件处理队列
  m_mapMoveObject = []
  m_endPageArr;
  m_showPetalFlag;
  m_elementPetalArr;
  m_showElementPetalFlag;
  m_PetalImage = "_scrap-pic-"  // 飘落动画
  m_renderArr // 渲染队列
  m_renderObject = [];
  m_defaultZindex = 0;
  m_setTimeoutIDs = [];
  m_setIntervalIDs = [];
  m_moveAsstantIntervalId = null;
  // ------------------------------------
  

  // ------------  游戏逻辑数据  ------------
  m_answerCard_All = []
  m_questionCard_All = []
  m_enableCardMove = false;
  m_currentSelectedCardID = null
  // ------------------------------------


  // ------------  消息  ------------
  
  // ------------------------------------


  // ------------  调试变量  ------------
  g_EnableStageRuler = false; // 使能舞台背景格尺
  g_ForceChangeDefaultRole = false // 强制当前角色为默认角色
  g_EnableTestSendEvent = false // 发送模拟Web数据
  g_EnableLeftCornerTest= false // 测试左上角图标
  // ------------------------------------


  // 当数据加载完毕后，执行
  systemReady(){
    if(this.g_EnableLeftCornerTest){
      this.setLeftCornerTest()
    }
    this.initGame()
  }
  
  // 屏幕尺寸变化后执行
  handleScreenResize(){
    this.initSystem();
    this.cleanSystemVar()
    this.cleanGameVar();
    if(this.g_EnableLeftCornerTest){
      this.setLeftCornerTest()
    }
    this.initGame()
  }

  // 映射预加载图片[网路]资源 返回包含图片路径的数组
  mapToImageArray(contentObj){
    let array = []
    this.g_formData.dataArray.forEach(element => {
      array.push(element.image_url)
    });
    return array
  }

  // 映射预加载音频[网路]资源 返回包含音频路径的数组
  mapToAduioArray(contentObj){
    let array = []
    return array
  }

  // ------------------------------------------------------------------------------
  // 游戏核心处理区
  // ------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------
  // 
  // 
  // 
  // ------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------

  initGame(){
    console.log(this.g_formData);
    this.initBackground();
    this.initCardsAll();
    this.initQuestionBackground()
    this.initQuestion();
  }

  cleanGameVar(){
    let audioProgress = this.g_cartoon.getCartoonElement("audio-plyer-progress")
    if(audioProgress && audioProgress.audio){
      audioProgress.audio.pause()
      audioProgress.audio.currentTime = 0;
      let slider = this.g_cartoon.getCartoonElement("audio-plyer-slider")
      if(slider.enableSlider){
        slider.release()
        this.disableMoveAsstant();
      }
    }
    this.m_currentSelectedCardID = null
    this.m_enableCardMove = false;
    this.m_answerCard_All = []
    this.m_questionCard_All = []
  }

  startGame(){

  }

  restartGame(){
    this.cleanGameVar()
    this.startGame()
  }

  endGame(){

  }

  initBackground(){
    let element = this.g_cartoon.createCartoonElementImageFunc("background-main", "bg_bg", (w, h)=>{
      let sx = this.g_canvasWidth / w
      let sy = this.g_canvasHeight / h
      let s = Math.min(sx, sy)
      return {
        sx: s,
        sy: s
      }
    }, (w, h)=>{
      return {
        x: this.g_canvasWidth / 2,
        y: this.g_canvasHeight / 2
      }
    })
    let xia = this.g_cartoon.createCartoonElementImage("background-main-xia", "bg_bottom", 1280 * element.ref.scaleX, 142 * element.ref.scaleY, this.g_canvasWidth / 2, this.g_cartoon.getOrigin().y + (720 - 71) * this.g_mapScale)

    this.subscribeMapUpEvent("background-color", ()=>{
      let slider = this.g_cartoon.getCartoonElement("audio-plyer-slider")
      if(slider && slider.enableSlider){
        slider.release()
        this.disableMoveAsstant();
      }
      if(this.m_enableCardMove && this.m_currentSelectedCardID){
        this.g_cartoon.getCartoonElement(this.m_currentSelectedCardID).release()
      }
      this.g_enableMapUp = true;
    })

    this.render(element.ref, zIndexMap.mainBackground)
    this.render(xia.ref, zIndexMap.mainBackground)
  }

  initCardsAll(){
    if(this.g_formData.answerType=="Text"){
      this.initAnsterCard_Text();
      this.initQuestionCard_Text();
    }else{
      this.initAnsterCard_Image();
      this.initQuestionCard_Image()
    }
  }

  // 问题卡片 文字
  initQuestionCard_Text(){
    this.g_formData.dataArray.forEach((element, index) => {
      this.m_questionCard_All.push(this.createQuestionCardText(index, element))
    });

    this.alignCenter(this.m_questionCard_All, 20*this.g_mapScale, true)
  }

  createQuestionCardText(index, text){

    let element = this.g_cartoon.createCartoonElementLabel("questioncard-text-"+index, " " + text + " ","BRLNSDB", "#FFFFFF", 60 * this.g_mapScale, 0, this.g_cartoon.getOrigin().y + 720 * this.g_mapScale - 250 * this.g_mapScale)
    let imageBD = element.ref.getBoundingBox()
    let cardHeight = 127
     
    // 边框
    let borderLeft  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-left-"+index, "image_border_left", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: (((cardHeight - 8) * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: -(imageBD.width / element.ref.scaleX) / 2,
        y: 13
      }
    })
    element.ref.addChild(borderLeft.ref)

    let borderRight  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-right-"+index, "image_border_right", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: (((cardHeight - 8) * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: (imageBD.width / element.ref.scaleX) / 2,
        y: 13
      }
    })
    element.ref.addChild(borderRight.ref)

    let border  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-"+index, "text_center", (w, h)=>{
      return{
        sx: (imageBD.width / element.ref.scaleX) / w,
        sy: ((cardHeight * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: 13
      }
    })
    element.ref.addChild(border.ref)

    // 高亮边框
    let border_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-h-"+index, "border_center", (w, h)=>{
      return{
        sx: (imageBD.width / element.ref.scaleX) / w,
        sy: ((cardHeight * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: 8
      }
    })
    element.ref.addChild(border_H.ref)

    let borderLeft_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-h-left-"+index, "border_left", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((cardHeight * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: -(imageBD.width / element.ref.scaleX) / 2 - 5,
        y: 9
      }
    })
    element.ref.addChild(borderLeft_H.ref)

    let borderRight_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-text-border-h-right-"+index, "border_right", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((cardHeight * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: (imageBD.width / element.ref.scaleX) / 2 + 5,
        y: 9
      }
    })
    element.ref.addChild(borderRight_H.ref)

    let text_copy = this.g_cartoon.createCartoonElementLabel("questioncard-text-"+index+"-copy", text,"BRLNSDB", "#d94158", 60 * this.g_mapScale, 0, 0)
    let text_copy_shadow = this.g_cartoon.createCartoonElementLabel("questioncard-text-"+index+"-copy", text,"BRLNSDB", "#fdf0cd", 60 * this.g_mapScale, 0, 5)
    element.ref.addChild(text_copy_shadow.ref)
    element.ref.addChild(text_copy.ref)

    border_H.ref.visible = false;
    borderRight_H.ref.visible = false;
    borderLeft_H.ref.visible = false;

    element.startPos = { x:element.initX, y:element.initY }
    element.backTo = (pos?, callback?)=>{
      if(pos){
        tweenChange(element.ref, pos, 0.1, ()=>{
          callback && callback()
        })
      }else{
        tweenChange(element.ref, { x:element.initX, y:element.initY }, 0.3, ()=>{
          callback && callback()
        })
      }
    }

    element.disable = false;
    element.release = (pos, callback?)=>{
      this.m_enableCardMove = false;
      this.m_currentSelectedCardID = null;
      border_H.ref.visible = false;
      borderRight_H.ref.visible = false;
      borderLeft_H.ref.visible = false;
      this.disableMoveAsstant()
      element.backTo(pos, callback)
    }

    this.subscribeMapDownEvent(element.id, ()=>{
      if(element.disable){
        this.g_enableMapDown = true;
        return
      }
      this.m_enableCardMove = true;
      this.m_currentSelectedCardID = element.id
      border_H.ref.visible = true;
      borderRight_H.ref.visible = true;
      borderLeft_H.ref.visible = true;
      this.topOfRenderArray(element.ref)
      element.startPos = {
        x: this.g_clickX,
        y: this.g_clickY
      }
      this.enableMoveAsstant(()=>{
        element.ref.x = this.g_clickX
        element.ref.y = this.g_clickY
      })
      this.g_enableMapDown = true;
    })

    this.subscribeMapUpEvent(element.id, ()=>{
      if(element.disable){
        this.g_enableMapUp = true;
        return
      }
      element.release()
      this.g_enableMapUp = true;
    })


    this.render(element.ref, zIndexMap.questionCard)
    return element
  }

  // 问题卡片 图片
  initQuestionCard_Image(){
    this.m_questionCard_All.push(this.createQuestionCardImage(0, {}))
    this.m_questionCard_All.push(this.createQuestionCardImage(1, {}))
    this.m_questionCard_All.push(this.createQuestionCardImage(2, {}))
    this.m_questionCard_All.push(this.createQuestionCardImage(3, {}))

    this.alignCenter(this.m_questionCard_All, 20*this.g_mapScale, true)
  }

  createQuestionCardImage(index, data){
    let element  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-"+index, "lego-building", (w, h)=>{
      let s = (155 * this.g_mapScale) / h
      return{
        sx: s,
        sy: s
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: this.g_cartoon.getOrigin().y + 720 * this.g_mapScale - 345 * this.g_mapScale
      }
    })
    let imageBD = element.ref.getBoundingBox()

    // 边框
    let border  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-"+index, "image_center", (w, h)=>{
      return{
        sx: (imageBD.width / element.ref.scaleX) / w,
        sy: ((imageBD.height + 26 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: 13
      }
    })
    element.ref.addChild(border.ref)

    let borderLeft  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-left-"+index, "image_border_left", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((imageBD.height + 26 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: -(imageBD.width / element.ref.scaleX) / 2 - 5,
        y: 13
      }
    })
    element.ref.addChild(borderLeft.ref)

    let borderRight  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-right-"+index, "image_border_right", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((imageBD.height + 26 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: (imageBD.width / element.ref.scaleX) / 2 + 5,
        y: 13
      }
    })
    element.ref.addChild(borderRight.ref)


    // 高亮边框
    let border_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-h-"+index, "border_center", (w, h)=>{
      return{
        sx: (imageBD.width / element.ref.scaleX) / w,
        sy: ((imageBD.height + 38 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: 0
      }
    })
    element.ref.addChild(border_H.ref)

    let borderLeft_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-h-left-"+index, "border_left", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((imageBD.height + 38 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: -(imageBD.width / element.ref.scaleX) / 2 - 20,
        y: 3
      }
    })
    element.ref.addChild(borderLeft_H.ref)

    let borderRight_H  = this.g_cartoon.createCartoonElementImageFunc("questioncard-image-border-h-right-"+index, "border_right", (w, h)=>{
      return{
        sx: (22 / element.ref.scaleX) / w,
        sy: ((imageBD.height + 38 * this.g_mapScale) / element.ref.scaleY) / h
      }
    }, (w, h)=>{
      return{
        x: (imageBD.width / element.ref.scaleX) / 2 + 20,
        y: 3
      }
    })
    element.ref.addChild(borderRight_H.ref)

    border_H.ref.visible = false;
    borderRight_H.ref.visible = false;
    borderLeft_H.ref.visible = false;

    element.startPos = { x:element.initX, y:element.initY }
    element.backTo = (pos?, callback?)=>{
      if(pos){
        tweenChange(element.ref, pos, 0.1, ()=>{
          callback && callback()
        })
      }else{
        tweenChange(element.ref, { x:element.initX, y:element.initY }, 0.3, ()=>{
          callback && callback()
        })
      }
    }

    element.disable = false;
    element.release = (pos, callback?)=>{
      this.m_enableCardMove = false;
      this.m_currentSelectedCardID = null;
      border_H.ref.visible = false;
      borderRight_H.ref.visible = false;
      borderLeft_H.ref.visible = false;
      this.disableMoveAsstant()
      element.backTo(pos, callback)
    }

    this.subscribeMapDownEvent(element.id, ()=>{
      if(element.disable){
        this.g_enableMapDown = true;
        return
      }
      this.m_enableCardMove = true;
      this.m_currentSelectedCardID = element.id
      border_H.ref.visible = true;
      borderRight_H.ref.visible = true;
      borderLeft_H.ref.visible = true;
      this.topOfRenderArray(element.ref)
      element.startPos = {
        x: this.g_clickX,
        y: this.g_clickY
      }
      this.enableMoveAsstant(()=>{
        element.ref.x = this.g_clickX
        element.ref.y = this.g_clickY
      })
      this.g_enableMapDown = true;
    })

    this.subscribeMapUpEvent(element.id, ()=>{
      if(element.disable){
        this.g_enableMapUp = true;
        return
      }
      element.release()
      this.g_enableMapUp = true;
    })


    this.render(element.ref, zIndexMap.questionCard)
    return element
  }

  // 答案卡片 文字
  initAnsterCard_Text(){
    this.m_answerCard_All.push(this.createAnswerCardText(0, {}))
    this.m_answerCard_All.push(this.createAnswerCardText(1, {}))
    this.m_answerCard_All.push(this.createAnswerCardText(2, {}))
    this.m_answerCard_All.push(this.createAnswerCardText(3, {}))

    this.alignCenter(this.m_answerCard_All, 10*this.g_mapScale, true)
  }

  createAnswerCardText(index, data){
    let element  = this.g_cartoon.createCartoonElementImageFunc("answercard-text-"+index, "bg_change", (w, h)=>{
      return{
        sx: (166 * this.g_mapScale) / w,
        sy: (128 * this.g_mapScale) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: this.g_cartoon.getOrigin().y + 720 * this.g_mapScale - 110 * this.g_mapScale
      }
    })

    this.subscribeMapUpEvent(element.id, ()=>{
      if(this.m_currentSelectedCardID && this.m_enableCardMove){
        let selectedCard = this.g_cartoon.getCartoonElement(this.m_currentSelectedCardID)
        selectedCard.ref.x = element.ref.x
        selectedCard.ref.y = element.ref.y
        selectedCard.release({x:element.ref.x, y:element.ref.y - 10 * this.g_mapScale}, ()=>{
          if(this.g_formData.answerType=="Image"){
            this.alignCenter(this.m_answerCard_All, 30*this.g_mapScale, true)
          }else{
            this.alignCenter(this.m_answerCard_All, 20*this.g_mapScale, true)
          }
        })
        selectedCard.disable = true;
        let index = this.m_answerCard_All.indexOf(element)
        this.m_answerCard_All.splice(index, 1, selectedCard)
        this.deleteElementInRender(element.id)
        this.topOfRenderArray(this.g_cartoon.getCartoonElement("background-main-xia").ref)
      }
      this.g_enableMapUp = true;
      return true
    }, 100)

    this.render(element.ref, zIndexMap.answerCard)
    return element
  }

  // 答案卡片 图片
  initAnsterCard_Image(){
    this.m_answerCard_All.push(this.createAnswerCardImage(0, {}))
    this.m_answerCard_All.push(this.createAnswerCardImage(1, {}))
    this.m_answerCard_All.push(this.createAnswerCardImage(2, {}))
    this.m_answerCard_All.push(this.createAnswerCardImage(3, {}))

    this.alignCenter(this.m_answerCard_All, 20*this.g_mapScale, true)
  }

  createAnswerCardImage(index, data){
    let element  = this.g_cartoon.createCartoonElementImageFunc("answercard-image-"+index, "bg_change2", (w, h)=>{
      return{
        sx: (166 * this.g_mapScale) / w,
        sy: (185 * this.g_mapScale) / h
      }
    }, (w, h)=>{
      return{
        x: 0,
        y: this.g_cartoon.getOrigin().y + 720 * this.g_mapScale - 140 * this.g_mapScale
      }
    })

    this.subscribeMapUpEvent(element.id, ()=>{
      if(this.m_currentSelectedCardID && this.m_enableCardMove){
        let selectedCard = this.g_cartoon.getCartoonElement(this.m_currentSelectedCardID)
        selectedCard.ref.x = element.ref.x
        selectedCard.ref.y = element.ref.y
        selectedCard.release({x:element.ref.x, y:element.ref.y - 6 * this.g_mapScale}, ()=>{
          if(this.g_formData.answerType=="Image"){
            this.alignCenter(this.m_answerCard_All, 30*this.g_mapScale, true)
          }else{
            this.alignCenter(this.m_answerCard_All, 20*this.g_mapScale, true)
          }
        })
        selectedCard.disable = true;
        let index = this.m_answerCard_All.indexOf(element)
        this.m_answerCard_All.splice(index, 1, selectedCard)
        this.deleteElementInRender(element.id)
        this.topOfRenderArray(this.g_cartoon.getCartoonElement("background-main-xia").ref)
      }
      this.g_enableMapUp = true;
      return true
    }, 100)

    this.render(element.ref, zIndexMap.answerCard)
    return element
  }

  initQuestionBackground(){
    if(this.g_formData.question.type=="Audio"){
      let element  = this.g_cartoon.createCartoonElementImageFunc("question-audio-bg", "bg_kuang2", (w, h)=>{
        return{
          sx: this.g_cartoon.getCartoonElement("background-main").ref.scaleX,
          sy: this.g_cartoon.getCartoonElement("background-main").ref.scaleY
        }
      }, (w, h)=>{
        return{
          x: this.g_canvasWidth / 2,
          y: this.g_cartoon.getOrigin().y + 127 * this.g_mapScale
        }
      })
  
      this.render(element.ref, zIndexMap.questionKuang)
    }else{
      let element  = this.g_cartoon.createCartoonElementImageFunc("question-text-image-bg", "bg_kuang1", (w, h)=>{
        return{
          sx: this.g_cartoon.getCartoonElement("background-main").ref.scaleX,
          sy: this.g_cartoon.getCartoonElement("background-main").ref.scaleY
        }
      }, (w, h)=>{
        return{
          x: this.g_canvasWidth / 2,
          y: this.g_cartoon.getOrigin().y + 225 * this.g_mapScale
        }
      })
  
      this.render(element.ref, zIndexMap.questionKuang)
    }
  }
  
  initQuestion(){
    let audio = this.initAudioPlayer()
    if(this.g_formData.question.type == "Audio"){
      // 进度条
      let progressBar = this.g_cartoon.createCartoonElementImageFunc("audio-plyer-progress", "bg_tiaodi", (w, h)=>{
        return {
          sx: 476 * this.g_mapScale / w,
          sy: 48 * this.g_mapScale / h
        }
      }, (w, h)=>{
        return {
          x: this.g_canvasWidth / 2 -  70 * this.g_mapScale,
          y: this.g_cartoon.getOrigin().y + 100 * this.g_mapScale
        }
      })
      progressBar.audio= audio

      // 时间标签
      let time = 0;
      if(audio){
        time = audio.duration
      }
      let endTime = Math.floor(time / 60)
      let timeString = ""
      if(endTime<9){
        timeString = "0" + endTime
      }else{
        timeString + endTime + ""
      }
      if(time % 60<9){
        timeString = timeString + ":0" + (time % 60 + "").substring(0, 2)
      }else{
        timeString = timeString + ":" + (time % 60 + "").substring(0, 2)
      }
      
      let timeStart = this.g_cartoon.createCartoonElementLabel("audio-player-start", "00:00", "BRLNSR_1", "#FFFFFF", 34)
      timeStart.ref.x = -(476/2) + 30
      timeStart.ref.y = 40
      progressBar.currentTimeString = timeStart

      let timeEnd = this.g_cartoon.createCartoonElementLabel("audio-player-start", timeString,  "BRLNSR_1", "#FFFFFF", 34)
      timeEnd.ref.x = (476/2) - 30
      timeEnd.ref.y = 40

      // 滑块
      let slider = this.g_cartoon.createCartoonElementImageFunc("audio-plyer-slider", "btn_tiaodian", (w, h)=>{
        return {
          sx: 54 / w,
          sy: 64 / h
        }
      }, (w, h)=>{
        return {
          x: -476/2,
          y: 0
        }
      })
      
      slider.enableSlider = false;
      slider.release = ()=>{
        if(!audio){
          this.g_enableMapUp = true;
          return;
        }
        slider.enableSlider= false;
        this.g_enableMapDown = true;
        progressBar.refreshAudioProgress();
      }
      progressBar.update = ()=>{
        coverColor.width = slider.ref.x - (-476/2)
      }

      // 已播放的覆盖颜色
      let coverColor = this.g_cartoon.createRectangula({
        width: 0,
        height: 38,
        x: -(476/2),
        y: -48/2,
        radius: 18,
        fillColor: "#f4ce48"
      })

      // 音频时间
      if(audio){
        audio.onended = ()=>{
          let audioPlay = this.g_cartoon.getCartoonElement("bg-short-audio") 
          audioPlay.playIcon.ref.visible = true
          audioPlay.pauseIcon.ref.visible = false
          audioPlay.audio.pause()
          audioPlay.isPlaying = !audioPlay.isPlaying

          slider.ref.x = -(476/2)
          coverColor.width = 0;
          progressBar.currentTimeString.ref.text = "00:00"
        }
        audio.ontimeupdate = ()=>{
          if(!slider.enableSlider){
            let per = audio.currentTime / time
            slider.ref.x = -(476/2) + 476 * per
            coverColor.width = slider.ref.x - (-(476/2))
          }
          let _time = parseInt(audio.currentTime)
          let m = parseInt(_time/60 + "")
          let s = parseInt(_time%60 + "")
          progressBar.currentTimeString.ref.text = `${m<9?"0"+m:m}:${s<9?"0"+s:s}`
        }
        progressBar.refreshAudioProgress = ()=>{
          slider.enableSlider = false;
          let dis = slider.ref.x + (476/2)
          let per = dis / 476
          audio.currentTime = audio.duration * per
        }
      }
      
      progressBar.ref.addChild(timeStart.ref)
      progressBar.ref.addChild(timeEnd.ref)
      progressBar.ref.addChild(coverColor)
      progressBar.ref.addChild(slider.ref)

      // 滑动滑块
      this.subscribeMapDownEvent(slider.id, ()=>{
        if(!audio){
          this.g_enableMapDown = true;
          return;
        }
        if(!slider.enableSlider){
          slider.enableSlider= true;
          this.enableMoveAsstant(()=>{
            let moveDis = (this.g_clickX - progressBar.ref.x) / this.g_mapScale
            if( moveDis > (476/2)){
              moveDis = (476/2)
            }else if( moveDis < (-476/2)){
              moveDis = -(476/2)
            }
            slider.ref.x = moveDis;
            progressBar.update()
          })
        }
        this.g_enableMapDown = true;
      })
      this.subscribeMapUpEvent(slider.id, ()=>{
        if(slider.enableSlider){
          slider.release()
          this.disableMoveAsstant();
        }
        this.g_enableMapUp = true;
      })

      this.render(progressBar.ref, zIndexMap.progressBar)
    }else{
      let element  = this.g_cartoon.createCartoonElementImageFunc("question-text-image-bg", "image_container", (w, h)=>{
        return{
          sx: this.g_cartoon.getCartoonElement("background-main").ref.scaleX,
          sy: this.g_cartoon.getCartoonElement("background-main").ref.scaleY
        }
      }, (w, h)=>{
        return{
          x: this.g_canvasWidth / 2,
          y: this.g_cartoon.getOrigin().y + 215 * this.g_mapScale
        }
      })

      let image = this.g_cartoon.createImage("lego-building", (w, h)=>{
        let sx = 590 / w
        let sy = 360 / h
        let s = Math.min(sx, sy)
        return {
          sx: s,
          sy: s
        }
      }, (w, h)=>{
        return {
          x: 0,
          y: 0
        }
      })
      element.ref.addChild(image)

      this.render(element.ref, zIndexMap.imageContainer)
    }

    
  }

  initAudioPlayer(){
    let element  = this.g_cartoon.createCartoonElementImageFunc("bg-short-audio", "bg_laba", (w, h)=>{
      return{
        sx: 474 * this.g_mapScale / w,
        sy: 98 * this.g_mapScale / h
      }
    }, (w, h)=>{
      let x = 0;
      let y = 0
      if(this.g_formData.question.type == "Audio"){
        x = -235
        y = -45
      }else{
        x = 20
      }
      return{
        x: this.g_cartoon.getOrigin().x + (1280 + x) * this.g_mapScale,
        y: this.g_cartoon.getOrigin().y + (220 + y) * this.g_mapScale
      }
    })

    element.playIcon = this.g_cartoon.createCartoonElementImageFunc("btn-audio-play", "btn_videooff", (w, h)=>{
      return{
        sx: 91 / w,
        sy: 90 / h
      }
    }, (w, h)=>{
      return{
        x: -140,
        y: -80
      }
    })
    element.pauseIcon = this.g_cartoon.createCartoonElementImageFunc("btn-audio-pause", "btn_videoon", (w, h)=>{
      return{
        sx: 91 / w,
        sy: 90 / h
      }
    }, (w, h)=>{
      return{
        x: -140,
        y: -80
      }
    })
    element.pauseIcon.ref.visible = false;

    element.ref.addChild(element.playIcon.ref)
    element.ref.addChild(element.pauseIcon.ref)
    
    element.audio = this.g_cartoon.audio.get("default_audio")
    element.isPlaying = false
    this.subscribeMapDownEvent(element.playIcon.id, ()=>{
      if(element.isPlaying){
        element.playIcon.ref.visible = true
        element.pauseIcon.ref.visible = false
        element.audio.pause()
        element.isPlaying = !element.isPlaying
      }else{
        element.playIcon.ref.visible = false
        element.pauseIcon.ref.visible = true
        element.audio.play()
        element.isPlaying = !element.isPlaying
      }
      this.g_enableMapDown = true;
    })

    this.render(element.ref, zIndexMap.shortAudio)

    return element.audio
  }

























































  // --------------------------------------------------
  // --------------  Template function  ---------------
  // --------------------------------------------------
  // --------------------------------------------------
  // --------------------------------------------------
  // 
  //                                  _________
  //                                 /         /.
  //           .-------------.      /_________/ |
  //         /             / |      |         | |
  //        /+============+\ |      | |====|  | |
  //        ||C:\>        || |      |         | |
  //        ||            || |      | |====|  | |
  //        ||            || |      |   ___   | |
  //        ||            || |      |  |166|  | |
  //        ||            ||/@@@    |   ---   | |
  //        \+============+/    @   |_________|./.
  //                          @          ..  ....'
  //        ..................@     __.'.'  ''
  //      /oooooooooooooooo//     ///
  //      /................//     /_/
  //      ------------------
  //
  // --------------------------------------------------

  @ViewChild("canvas") canvas: ElementRef;
  @ViewChild("wrap") wrap: ElementRef;
  @HostListener("window:resize", ["$event"])
  onResize(event) {
    this.g_winResizeEventStream.next();
  }
  ngOnDestroy() {
    window["curCtx"] = null;
    window.cancelAnimationFrame(this.g_animationId);
  }
  ngOnInit() {
    const getData = (<any>window).courseware.getData;
    getData(data => {
      if (window['air'].airClassInfo.user.classRole == 'tea') {
        if(!this.g_ForceChangeDefaultRole){
          this.g_teacherFlag = true;
        }
      }
      this.g_currentUser = window['air'].airClassInfo.user;
      if (data && typeof data == "object") {
        this.g_data = data;
        this.g_formData = data.contentObj;
      } else {
        this.g_data = {};
      }

      
      if (!this.g_data.contentObj) {
        this.g_data.contentObj = {};
        this.g_formData = {};
      }

      this.initDefaultData();
      this.initAudio();
      this.initImg();

       // 预加载资源
       this.g_cartoon.loadResources().then(() => {
        window["air"].hideAirClassLoading(this.g_KEY, this.g_data);
        this.initSystem();
        this.update();
        this.systemReady()
      });

      this.initListener();
    }, this.g_KEY);
  }

  // ----------------------------------
  // 初始化默认数据
  // ----------------------------------
  initDefaultData() {
    if ( Object.keys(this.g_formData).length===0 ) {
      this.g_formData = defauleFormData;
    }
  }


  // ----------------------------------
  // 初始化音乐
  // ----------------------------------
  initAudio() {
    const contentObj = this.g_formData;
    if (!contentObj) {
      return;
    }
    // 添加用户上传音效
    let images:Array<string> = this.mapToAduioArray(contentObj)
    images.forEach(image => {
      this.g_cartoon.addAudio( image, image );
    });

    // 添加本地音效
    for( var key in localAudios ){
      this.g_cartoon.addAudio( key, localAudios[key] );
    }
  }

  // ----------------------------------
  // 初始化图片
  // ----------------------------------
  initImg() {
    const contentObj = this.g_formData;
    if (contentObj) {
      const addPicUrl = url => {
        if (url) {
          this.g_cartoon.addImage(url, url);
        }
      };
      let images:Array<string> = this.mapToImageArray(contentObj)
      images.forEach(image => {
        addPicUrl(image)
      });
    }
    // 添加本地图片
    for( var key in localImages ){
     this.g_cartoon.addImage( key, localImages[key] );
    }
  }


  mapDown(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapDown) {
      return;
    }

    this.m_mapDownArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapDown = false;
          if(item.callback()){
            myStopPropagation = true;
          }
        }
      }
    })


  }

  mapMove(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapMove) {
      return;
    }

    this.m_mapMoveArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapMove = false;
          if(item.callback()){
            myStopPropagation = true;
          }
        }
      }
    })
  }

  mapUp(event) {
    let myStopPropagation = false;
    if (!this.g_enableMapUp) {
      return;
    }
    // for(let cartoonID in this.m_mapUpQueue){
    //   if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(cartoonID))) {
    //     this.g_enableMapUp = false;
    //     this.m_mapUpQueue[cartoonID]();
    //   }
    // }
    this.m_mapUpArray.forEach((item)=>{
      if(!myStopPropagation){
        if (this.checkClickTarget(this.g_cartoon.getCartoonElementRef(item.id))) {
          this.g_enableMapUp = false;
          if(item.callback()){
            myStopPropagation = true;
          }
        }
      }
    })

  }


  update() {
    this.g_animationId = window.requestAnimationFrame(this.update.bind(this));
    // 清除画布内容
    this.g_ctx.clearRect(0, 0, this.g_canvasWidth, this.g_canvasHeight);
    TWEEN.update();
    this.updateArr(this.m_renderArr);
    this.updateArr(this.m_endPageArr);
    this.updateArr(this.m_elementPetalArr);
  }


  updateItem(item) {
    if (item) {
      item.update();
    }
  }


  updateArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].update(this);
    }
  }


  initListener() {
    const element = this.canvas.nativeElement;
    this.g_winResizeEventStream.pipe(debounceTime(500)).subscribe(data => {
      this.renderAfterResize();
    });
    
    const addTouchListener = () => {
      element.addEventListener('touchstart', touchDownFunc);
      element.addEventListener('touchmove', touchMoveFunc);
      element.addEventListener('touchend', touchUpFunc);
      element.addEventListener('touchcancel', touchUpFunc);
    };
    const removeTouchListener = () => {
      element.removeEventListener('touchstart', touchDownFunc);
      element.removeEventListener('touchmove', touchMoveFunc);
      element.removeEventListener('touchend', touchUpFunc);
      element.removeEventListener('touchcancel', touchUpFunc);
    };

    const addMouseListener = () => {
      element.addEventListener('mousedown', mouseDownFunc);
      element.addEventListener('mousemove', mouseMoveFunc);
      element.addEventListener('mouseup', mouseUpFunc);
    };
    const removeMouseListener = () => {
      element.removeEventListener('mousedown', mouseDownFunc);
      element.removeEventListener('mousemove', mouseMoveFunc);
      element.removeEventListener('mouseup', mouseUpFunc);
    };

    const touchDownFunc = (e) => {
      if (this.g_firstTouch) {
        this.g_firstTouch = false;
        removeMouseListener();
      }
      setMxMyByTouch(e);
      this.mapDown(e);
    };
    const touchMoveFunc = (e) => {
      setMxMyByTouch(e);
      this.mapMove(e);
    };
    const touchUpFunc = (e) => {
      setMxMyByTouch(e);
      this.mapUp(e);
    };

    const mouseDownFunc = (e) => {
      if (this.g_firstTouch) {
        this.g_firstTouch = false;
        removeTouchListener();
      }
      setMxMyByMouse(e);
      this.mapDown(e);
    };
    const mouseMoveFunc = (e) => {
      setMxMyByMouse(e);
      this.mapMove(e);
    };
    const mouseUpFunc = (e) => {
      setMxMyByMouse(e);
      this.mapUp(e);
    };

    const setMxMyByTouch = event => {
      if (event.touches.length <= 0) {
        return;
      }
      if (this.g_canvasLeft == null) {
        setParentOffset();
      }
      this.g_clickX = event.touches[0].pageX - this.g_canvasLeft;
      this.g_clickY = event.touches[0].pageY - this.g_canvasTop;
    };

    const setParentOffset = () => {
      const rect = this.canvas.nativeElement.getBoundingClientRect();
      this.g_canvasLeft = rect.left;
      this.g_canvasTop = rect.top;
    };

    const setMxMyByMouse = (event) => {
      this.g_clickX = event.offsetX;
      this.g_clickY = event.offsetY;
    };

    addMouseListener();
    addTouchListener();
  }

  showArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].visible = true;
    }
  }

  hideArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].visible = false;
    }
  }

  IsPC() {
    if (window["ELECTRON"]) {
      return false; // 封装客户端标记
    }
    if (
      document.body.ontouchmove !== undefined &&
      document.body.ontouchmove !== undefined
    ) {
      return false;
    } else {
      return true;
    }
  }

  renderAfterResize() {
    this.g_canvasWidth = this.wrap.nativeElement.clientWidth;
    this.g_canvasHeight = this.wrap.nativeElement.clientHeight;
    this.update();
    this.handleScreenResize()
  }

  checkClickTarget(target) {
    if (!target) {
      return false;
    }
    const rect = target.getBoundingBox();
    if (this.checkPointInRect(this.g_clickX, this.g_clickY, rect)) {
      return true;
    }
    return false;
  }

  getWorlRect(target) {
    let rect = target.getBoundingBox();
    if (target.parent) {
      const pRect = this.getWorlRect(target.parent);
      rect.x += pRect.x;
      rect.y += pRect.y;
    }
    return rect;
  }

  checkPointInRect(x, y, rect) {
    if (x >= rect.x && x <= rect.x + rect.width) {
      if (y >= rect.y && y <= rect.y + rect.height) {
        return true;
      }
    }
    return false;
  }


  getPosByAngle(angle, len) {
    const radian = (angle * Math.PI) / 180;
    const x = Math.sin(radian) * len;
    const y = Math.cos(radian) * len;
    return { x, y };
  }


  getPosDistance(sx, sy, ex, ey) {
    const _x = ex - sx;
    const _y = ey - sy;
    const len = Math.sqrt(Math.pow(_x, 2) + Math.pow(_y, 2));
    return len;
  }


  subscribeMapDownEvent(id,callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapDownObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapDownObject.sort((a,b)=>{
      return b.zIndex-a.zIndex
    })
    this.m_mapDownArray = []
    this.m_mapDownObject.forEach(item=>{
      this.m_mapDownArray.push(item)
    })
  }

  subscribeMapUpEvent(id,callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapUpObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapUpObject.sort((a,b)=>{
      return b.zIndex-a.zIndex
    })
    this.m_mapUpArray = []
    this.m_mapUpObject.forEach(item=>{
      this.m_mapUpArray.push(item)
    })
  }

  subscribeMapMoveEvent(id, callback, zIndex?){
    zIndex = zIndex?zIndex:0
    this.m_mapMoveObject.push({id:id, zIndex:zIndex?zIndex:1, callback:callback})
    this.m_mapMoveObject.sort((a,b)=>{
      return a.zIndex - b.zIndex
    })
    this.m_mapMoveArray = []
    this.m_mapMoveObject.forEach(item=>{
      this.m_mapMoveArray.push(item)
    })
  }


  // 全局初始化入口，当资源加载完毕后执行
  initSystem() {
    
    this.g_canvasWidth = this.wrap.nativeElement.clientWidth;
    this.g_canvasHeight = this.wrap.nativeElement.clientHeight;

    const sx = this.g_canvasWidth / this.g_canvasBaseW;
    const sy = this.g_canvasHeight / this.g_canvasBaseH;
    const s = Math.min(sx, sy);

    this.g_mapScale = s;
    this.g_cartoon.mapScale = this.g_mapScale;
    this.g_cartoon.clientWidth = this.wrap.nativeElement.clientWidth;
    this.g_cartoon.clientHeight = this.wrap.nativeElement.clientHeight;

    this.m_renderArr = [];
    this.m_renderObject = [];
    this.g_enableMapDown = true;

    this.g_ctx = this.canvas.nativeElement.getContext("2d");
    this.canvas.nativeElement.width = this.g_canvasWidth;
    this.canvas.nativeElement.height = this.g_canvasHeight;
    window["curCtx"] = this.g_ctx;

    // 初始化舞台
    this.initStage();
    this.update();
  }
  

  initStage() {
    const bgWidth = 1280;
    const bgHeight = 720;
    this.g_cartoon.stageWidth = bgWidth*this.g_mapScale;
    this.g_cartoon.stageHeight= bgHeight*this.g_mapScale;

    const imageColor = this.g_cartoon.createCartoonElement("background-color", "ShapeRect").ref;
    // console.log( "stageWidth: " + this.g_cartoon.stageWidth + " stageHeight" + this.g_cartoon.stageHeight)
    // console.log( "clientWidth: " + this.g_cartoon.clientWidth + " clientHeight" + this.g_cartoon.clientHeight)
    // console.log( "canvasWidth: " + this.g_canvasWidth + " canvasHeight" + this.g_canvasHeight)
    // console.log( "mapScale: " + this.g_mapScale )
    imageColor.x = this.g_cartoon.clientWidth / 2 ;
    imageColor.y = this.g_cartoon.clientHeight / 2 ;
    imageColor.setSize(this.g_cartoon.clientWidth, this.g_cartoon.clientHeight);
    imageColor.init();
    imageColor.fillColor = this.g_background_color;
    this.render(imageColor, -999999)

    const image = this.g_cartoon.createCartoonElement("bg_1280_720_Ruler", "MySprite").ref;
    image.init(this.g_cartoon.images.get("_bg_1280_720_Ruler"));
    image.x = this.g_canvasWidth / 2;
    image.y = this.g_canvasHeight /2;
    image.visible = this.g_EnableStageRuler
    this.g_cartoon.setOrigin( image.x-(bgWidth/2*this.g_mapScale), image.y-(bgHeight/2*this.g_mapScale)  )
    this.g_cartoon.setRelativeOrigin( -bgWidth/2, -bgHeight/2)
    image.setScaleXY(this.g_mapScale);
    this.render(image, 50)
    this.g_stage = image
  }

  render(ele, zIndex?:number){
    this.m_renderObject.push({element:ele, zIndex:zIndex?zIndex:1})
    this.m_renderObject.sort((a,b)=>{
      return a.zIndex - b.zIndex
    })
    this.m_renderArr = []
    this.m_renderObject.forEach(item=>{
      this.m_renderArr.push(item.element)
    })    
  }

  sendServerEvent(key, data) {
    const c = (<any> window).courseware;
    c.sendEvent(key, JSON.stringify(data));
  }

  addServerListener(msg_key, callback) {
    const c = (<any> window).courseware;
    c.onEvent(msg_key, (data,next) => {
      callback(JSON.parse(data))
      next && next();
    });
  }

  randomArray_shuffle(array) {
    var input = array;
    for (var i = input.length-1; i >=0; i--) {
        var randomIndex = Math.floor(Math.random()*(i+1));
        var itemAtIndex = input[randomIndex];
        input[randomIndex] = input[i];
        input[i] = itemAtIndex;
    }
    return input;
  }

  paginationArray(pageNo, pageSize, array) {
    var offset = (pageNo - 1) * pageSize;
    return (offset + pageSize >= array.length) ? array.slice(offset, array.length) : array.slice(offset, offset + pageSize);
  }

  stopAllTimeout(){
    this.m_setTimeoutIDs.forEach(id=>clearTimeout(id))
    this.m_setTimeoutIDs = []
  }

  stopAllInterval(){
    this.m_setIntervalIDs.forEach(id=>clearInterval(id))
    this.m_setIntervalIDs = []
  }

  topOfRenderArray(element){
    let index = this.m_renderArr.indexOf(element)
    if(index !=-1){
      this.m_renderArr.splice(index, 1)
      this.m_renderArr.push(element)
    }   
  }

  setRenderZIndex(element, zIndex){
    let index = null;
    for(let i=0; i<this.m_renderObject.length; i++){
      if( this.m_renderObject[i].element.id == element.id ){
        index = i
      }
    }
    if(index){
      this.m_renderObject.splice(index, 1)
      this.m_renderObject.push({element:element, zIndex:zIndex?zIndex:1})
      this.m_renderObject.sort((a,b)=>{
        return a.zIndex - b.zIndex
      })
  
      this.m_renderArr = []
      this.m_renderObject.forEach(item=>{
        this.m_renderArr.push(item.element)
      })
    }
  }
  
  deleteElementInRender(id){
    let index = null;
    for(let i=0; i<this.m_renderObject.length; i++){
      if( this.m_renderObject[i].element.id == id ){
        index = i
      }
    }
    if(index){
      this.m_renderObject.splice(index, 1)
      this.m_renderObject.sort((a,b)=>{
        return a.zIndex - b.zIndex
      })
      this.m_renderArr = []
      this.m_renderObject.forEach(item=>{
        this.m_renderArr.push(item.element)
      })
    }
  }

  enableMoveAsstant(callback){
    if(this.m_moveAsstantIntervalId){
      clearInterval(this.m_moveAsstantIntervalId)
    }
    this.m_moveAsstantIntervalId = setInterval(()=>{
      callback()
    },50)
  }

  disableMoveAsstant(){
    clearInterval(this.m_moveAsstantIntervalId)
  }

  cleanSystemVar(){
    this.m_mapDownQueue = {}
    this.m_mapDownArray = []
    this.m_mapDownObject = []
    this.m_mapMoveArray = []
    this.m_mapMoveObject = []
    this.m_mapUpQueue = {}
    this.m_mapUpArray = []
    this.m_mapUpObject = []
    this.stopAllInterval();
    this.stopAllTimeout()
    this.g_cartoon.stopAllAudio()
  }

  getMaxSubstringLength(str){
    let maxLength = 0;
    let subSubstring = str.split(" ")
    for (let index=0; index<subSubstring.length; index++) {
      if(subSubstring[index].length > maxLength){
        maxLength = subSubstring[index].length;
      }
    }
    return maxLength
  }

  setLeftCornerTest(){
    const bgRect = new ShapeRect();
    bgRect.setSize(57, 65);
    bgRect.fillColor = '#f8c224';
    const sx = this.g_canvasWidth / this.g_canvasBaseW;
    bgRect.setScaleXY(sx);
    bgRect.x = 65 * sx;
    this.g_partTitle_x = bgRect.x
    this.g_mainTitle_x = bgRect.x + 80 * sx
    bgRect.alpha = 0.5
    this.render(bgRect, 9999);
  }

  alignCenter(elementArray, spacing, withAni?, callback?){
    let totlaWidth = 0
    elementArray.forEach(element => {
      let bd = element.ref.getBoundingBox()
      totlaWidth += bd.width
    });
    totlaWidth += (elementArray.length-1) * spacing
    let laseX = this.g_canvasWidth / 2 - totlaWidth/2;
    elementArray.forEach(element => {
      let bd = element.ref.getBoundingBox()
      if(withAni){
        tweenChange(element.ref, {x: laseX + bd.width / 2}, 0.2, ()=>{
          this.g_cartoon.saveSize(element.id)
          callback && callback()
        })
      }else{
        element.ref.x = laseX + bd.width / 2
        this.g_cartoon.saveSize(element.id)
        callback && callback()
      }

      laseX = (laseX + bd.width + spacing)
    });
  }

  createTestLine(x, y, color?){
    var colorAll = ['#ff0000','#eb4310','#f6941d','#fbb417','#ffff00','#cdd541','#99cc33','#3f9337','#219167','#239676','#24998d','#1f9baa','#0080ff','#3366cc','#333399','#003366','#800080','#a1488e','#c71585','#bd2158'];
    color = color ? color : colorAll[Math.floor(Math.random()*20)]
    this.render(this.g_cartoon.createRectangula({
      width: 5,
      height: 100,
      x: x,
      y: y,
      fillColor: color
    }), 9999)
  }













  













  












  // --------------------------------------------------
  // --------------  Template function  ---------------
  // --------------------------------------------------
  // --------------------------------------------------
  // --------------------------------------------------
  // 
  //                .-~~~~~~~~~-._       _.-~~~~~~~~~-.
  //            __.'              ~.   .~              `.__
  //          .'//                  \./                  \\`.
  //        .'//                     |                     \\`.
  //      .'// .-~"""""""~~~~-._     |     _,-~~~~"""""""~-. \\`.
  //    .'//.-"                 `-.  |  .-'                 "-.\\`.
  //  .'//______.============-..   \ | /   ..-============.______\\`.
  //.'______________________________\|/______________________________`.
  //
  // --------------------------------------------------
  // showParticle( element_id )                 泡泡效果
  // --------------------------------------------------
  // showEndPatal()  /  stopEndPatal()  花瓣飘落结束动画 
  // --------------------------------------------------
  // showCorrectPatal()                 指定元素上面飘花
  // --------------------------------------------------
  // convertPercentToRadian()   将百分比转换为弧长 第一个参数是百分比，第二个参数是方向 true为逆时针，false为顺时针。用于ShapeCircle换圆弧
  // --------------------------------------------------
  // movePaoWuxian()                        元素抛物线跳
  // --------------------------------------------------
  // showJellyAnimation()------------------------------
  // showBlingStar()----------------------显示星星效果--
  // --------------------------------------------------
  


  // 泡泡
  showParticle(card) {
    let myCard = this.g_cartoon.getCartoonElementRelativePosition(card.id)
    showPopParticle(this.g_cartoon.images.get("_bubble"), { x: myCard.x , y: myCard.y }, this.g_stage);
  }

  // 选择正确动画
  showCorrectPatal(card_id, showTime, callback?) {
    this.m_elementPetalArr = [];
    this.m_showElementPetalFlag = true;
    this.addCorrectPetal(card_id);
    setTimeout(()=>{
      this.m_elementPetalArr = [];
      this.m_showElementPetalFlag = false;
      callback && callback()
    },showTime)
  }
  stopAllCorrectPatal(){
    this.m_elementPetalArr = [];
    this.m_showElementPetalFlag = false;
  }

  addCorrectPetal(card_id) {
    if (!this.m_showElementPetalFlag) {
      return;
    }

    let element = this.g_cartoon.getCartoonElement(card_id)
    const petal = new MySprite(this.g_ctx);

    const id = Math.ceil(Math.random() * 3);
    petal.init(this.g_cartoon.images.get(this.m_PetalImage + id));

    const randomS = (Math.random() * 0.4 + 0.6) * this.g_mapScale * 0.5;
    petal.setScaleXY(randomS);

    const randomR = Math.random() * 360;
    petal.rotation = randomR;

    const randomX = Math.random() * element.ref.width * this.g_mapScale;

    petal.x = element.ref.x - (element.ref.width * this.g_mapScale / 2) + randomX;
    petal.y = element.ref.y - (element.ref.height * this.g_mapScale / 2);

    const randomT = 1 + Math.random() * 2;
    petal["time"] = randomT;

    let randomTR = 360 * Math.random(); // - 180;
    if (Math.random() < 0.5) {
      randomTR *= -1;
    }
    petal["tr"] = randomTR;

    this.m_elementPetalArr.push(petal);
    moveItem(
      petal,
      petal.x,
      element.ref.y + (element.ref.height * this.g_mapScale / 2),
      petal["time"],
      () => {
        removeItemFromArr(this.m_elementPetalArr, petal);
      }
    );
    rotateItem(petal, petal["tr"], petal["time"]);
    setTimeout(() => {
      this.addCorrectPetal(card_id);
    }, 200);
  }


  // 结束动画花瓣飘落
  showEndPatal() {
    this.m_endPageArr = [];
    this.m_showPetalFlag = true;
    this.addPetal();
  }

  stopEndPatal() {
    this.m_endPageArr = [];
    this.m_showPetalFlag = false;
  }

  addPetal() {
    if (!this.m_showPetalFlag) {
      return;
    }
    const petal = this.getPetal();
    this.m_endPageArr.push(petal);
    moveItem(
      petal,
      petal.x,
      this.g_canvasHeight + petal.height * petal.scaleY,
      petal["time"],
      () => {
        removeItemFromArr(this.m_endPageArr, petal);
      }
    );
    rotateItem(petal, petal["tr"], petal["time"]);
    setTimeout(() => {
      this.addPetal();
    }, 100);
  }

  getPetal() {
    const petal = new MySprite(this.g_ctx);

    const id = Math.ceil(Math.random() * 3);
    petal.init(this.g_cartoon.images.get(this.m_PetalImage + id));

    const randomS = (Math.random() * 0.4 + 0.6) * this.g_mapScale;
    petal.setScaleXY(randomS);

    const randomR = Math.random() * 360;
    petal.rotation = randomR;

    const randomX = Math.random() * this.g_canvasWidth;
    petal.x = randomX;
    petal.y = (-petal.height / 2) * petal.scaleY;

    const randomT = 2 + Math.random() * 5;
    petal["time"] = randomT;

    let randomTR = 360 * Math.random(); // - 180;
    if (Math.random() < 0.5) {
      randomTR *= -1;
    }
    petal["tr"] = randomTR;

    return petal;
  }


  showJellyAnimation(element_id, callback?){
    let element = this.g_cartoon.getCartoonElement(element_id).ref
    tweenChange(element,{ scaleX: this.g_mapScale * 1.2 , scaleY: this.g_mapScale * 1.1 }, 0.1, ()=>{
      tweenChange(element,{ scaleX: this.g_mapScale * 0.8 , scaleY: this.g_mapScale * 0.9 }, 0.1, ()=>{
        tweenChange(element,{ scaleX: this.g_mapScale , scaleY: this.g_mapScale }, 0.1, ()=>{
          callback && callback()
        })
      })
    })
  }

  showShakeAnimation(element_id, callback?){
    let element = this.g_cartoon.getCartoonElement(element_id).ref
    let originX = element.x
    tweenChange(element,{ x: originX - 10*this.g_mapScale }, 0.1, ()=>{
      tweenChange(element,{ x: originX + 10*this.g_mapScale }, 0.1, ()=>{
        tweenChange(element,{ x: originX }, 0.1, ()=>{
          callback && callback()
        })
      })
    })
  }

  convertPercentToRadian(per, direction){
    let radian = 0;
    if(direction) { //逆时针
      if(per*360 > 270){
        radian = 361 - (per*360 - 270)
      }else{
        radian = 270 - per*360
      }
    }else{ //顺时针
      if(per*360 > 90){
        radian = per*360 - 90
      }else{
        radian = 270 + per*360
      }
    }
    return (radian * Math.PI) / 180;
  }

  showBlingStar(element, callback?){
    let rect =  element.getBoundingBox()
    showBlingBling(this.g_cartoon.images.get('icon_finishstar'), rect, element, 0.5, 1, 0.08, 2);
    this.m_setTimeoutIDs.push(setTimeout(() => {
      callback && callback()
    }, 2000))
    // setTimeout(()=>{
    //   showBlingBling(this.g_cartoon.images.get('icon_star'), rect, element, 0.5, 1, 0.08, 1.8);
    // },200)
    // setTimeout(()=>{
    //   showBlingBling(this.g_cartoon.images.get('icon_star'), rect, element, 0.5, 1, 0.08, 1.6);
    // },400)
  }

  movePaoWuxian(element,dis,time,callback?){
    let disX = dis.x - element.x
    let disY = (element.y - dis.y) 
    let count = 0
    let runTime = time/5
    let vx = disX/runTime
    let a = -0.05
    let vy0 = disY/runTime - 0.5*a*runTime
    let id = setInterval(()=>{
      element.x += vx
      element.y -= vy0 + a*count
      count++;
      if(count>runTime){
        clearInterval(id)
        callback && callback()
      }
    }, 5);
  }
}

